"""Plot data. There is only one main function plot. Check 'plot' documentation for how to use it."""

from mypythontools.plots.plots_internal import get_plotly_layout, plot

__all__ = ["get_plotly_layout", "plot"]
